#include <string.h>
#include <stdlib.h>
#include "error.h"
#include "config.h"
#include "reffile.h"
#include "loadfile.h"
#include "cdbslib.h"
#include "dbquery.h"
#include "hashtab.h"
#include "mkload.h"
#include "cdbserror.h"

#define  NONE		0
#define  ONCE		1
#define  MANY		2

static void		parse_pedigree (char *, char *, char *, char *, int);

//=========================================================================
// MKLOAD_BODY -- Write the body of the loadfile table
// 
// u: tab		loadfile table descriptor
// i: reffile_name      reference file name
// i: instrument 	instrument name
// i: reftype		reference file type
// i: filetype          image or table
//
//=========================================================================
void mkload_body (table * tab, 
                  char *reffile_name, 
                  char *instrument, 
                  char *reftype,
                  char *filetype)
{
    hashtab     *hash;
    int		irow, jrow, nrow, ikey, pedigree_instances;
    char	*keyword, *keyval;
    char	source[SZ_STRING+1];
    char	start_date[SZ_STRING+1], end_date[SZ_STRING+1];

    // Create non-obsmode columns to be filled later 
    static char change_level_colname[] = "CHANGE_LEVEL";
    static char	pedigree_colname[]     = "PEDIGREE";
    static char obstart_colname[]      = "OBSERVATION_BEGIN_DATE";
    static char obsend_colname[]       = "OBSERVATION_END_DATE";
    static char comment_colname[]      = "COMMENT";
    add_tab_col (tab, change_level_colname);
    add_tab_col (tab, pedigree_colname);
    add_tab_col (tab, obstart_colname);
    add_tab_col (tab, obsend_colname);
    add_tab_col (tab, comment_colname);

    // Read list of observation mode fields from the cdbs_mode DB table
    //   (will include both file selection and row selection fields)
    //
    textbuf obsmode;
    obsmode_qry (instrument, reftype, &obsmode);

    // open the primaryHDU of the reference file
    reffile *ref = open_reffile_hdr (reffile_name, filetype, "r");
    if (ref == NULL) {
        fprintf(stderr,"Failed to open %s as type=%s\n",
                        reffile_name, filetype);
        cdbs_warn (NOTOPEN, "%s", reffile_name);
        return;
    }

    // Search primary header for file selection observation mode values 
    textbuf obsvals;
    read_header (ref, &obsmode, &obsvals);

    // Grab the header pedigree keyword, if present
    char pedigree_hdr_val[SZ_STRING+1];
    pedigree_hdr_val[0] = '\0';
    get_hdr_reffile (ref, pedigree_colname, pedigree_hdr_val, SZ_STRING);

    // Determine file type
    char actual_file_type[20];
    type_reffile(reffile_name, actual_file_type);

    // IMAGE reffile
    if (strcmp(actual_file_type,"image")==0) {

        // No row select fields for images, so write observation mode keywords 
	for (ikey = 0; (keyword = get_text (&obsmode, ikey)) != NULL; ikey ++){
	    keyval = get_text (&obsvals, ikey);
	    if (keyval[0] != '\n') {
		put_tab_cell (tab, keyword, 0, keyval);
	    } else {
	        cdbs_error(NOHDRKEY, 
                           "MISSING IMAGE FILE SELECT keyword value: %s:%s",
                           name_reffile(ref), keyword);
	    }
	}

	// Write pedigree keywords 
	if (strlen(pedigree_hdr_val) > 0) {
	    strucase (pedigree_hdr_val);

	    parse_pedigree (pedigree_hdr_val, source, start_date, end_date, SZ_STRING);

	    put_tab_cell (tab, pedigree_colname, 0, source);
	    put_tab_cell (tab, obstart_colname, 0, start_date);
	    put_tab_cell (tab, obsend_colname, 0, end_date);
	}

    // TABLE reffile
    } else {
        // close primaryHDU and open first extension
        close_reffile(ref);
        ref = open_reffile_body (reffile_name, actual_file_type, "r");
        if (ref == NULL) {
           fprintf(stderr,"Failed to open %s as type=%s\n",
                           reffile_name, actual_file_type);
           cdbs_warn (NOTOPEN, "%s", reffile_name);
           return;
        }

        // see if the pedigree keyword is a TABLE column
        char notUsed[SZ_STRING+1];
        if (get_body_reffile (ref, pedigree_colname, 0, notUsed, SZ_STRING) != ERR) {
	    pedigree_instances = MANY;
	    
        // OR see if the pedigree keyword is a HEADER keyword
	} else if (strlen(pedigree_hdr_val) > 0) {
	    parse_pedigree (pedigree_hdr_val, source, start_date, end_date, SZ_STRING);
	    pedigree_instances = ONCE;

	} else {
	    pedigree_instances = NONE;
	}

	// Copy the info to each row of the table 
        jrow = 0;
	nrow = nrow_reffile (ref);
        hash = init_hash (nrow);

        char value[SZ_STRING+1];
	for (irow = 0; irow < nrow; irow ++) {

	    // Observation mode 
	    for (ikey = 0; 
		 (keyword = get_text (&obsmode, ikey)) != NULL; 
		 ikey ++) {

	        keyval = get_text (&obsvals, ikey);

		if (get_body_reffile (ref, keyword, irow, 
				      value, SZ_STRING) != ERR) {
                    // fill row select keywords from table content
		    put_tab_cell (tab, keyword, jrow, value);

		} else if (keyval[0] != '\n') {
                    // not a row select field, must be a file select field
                    // so fill keyword from saved header value
		    put_tab_cell (tab, keyword, jrow, keyval);

		} else {
		    cdbs_error (NOHDRKEY, 
                                "MISSING TABLE keyword value: %s:%s",
				name_reffile(ref), keyword);
		}
	    }  // end for ikey

	    // Pedigree 
	    if (pedigree_instances == MANY) {
	        get_body_reffile (ref, pedigree_colname, irow, value, SZ_STRING);
		parse_pedigree (value, source, start_date, 
				end_date, SZ_STRING);
	    }

	    if (pedigree_instances != NONE) {
	        put_tab_cell (tab, pedigree_colname, jrow, source);
		put_tab_cell (tab, obstart_colname, jrow, start_date);
		put_tab_cell (tab, obsend_colname, jrow, end_date);
	    }

	    // Check if observation mode is unique 
	    if (find_hash (hash, &obsmode, tab, jrow) == NO) {
	        jrow ++;
	    } else if (irow == nrow - 1) {
	        trunc_table (tab, jrow);
	    }
	} // end for irow

	free_hash (hash);
    }  // end else TABLE reffile

    free_text (&obsmode);
    free_text (&obsvals);

    close_reffile(ref);
}

//=====================================================================
// PARSE_PEDIGREE -- Break the pedigree keywords up into its fields
//
// i: pedigree		the value of the pedigree keyword
// o: field		its component fields
// i: nfield		the number of fields
//
//=====================================================================

static void parse_pedigree (char *pedigree, char *source, char *start_date, 
			    char *end_date, int maxch)
{
    char 	*dash, *field[3];

    if ((dash = strchr (pedigree, '-')) != NULL)
        *dash = ' ';

    strsplit (' ', pedigree, field, 3);
    if (field[2][0] == EOS)
        field[2] = field[1];

    safecopy (source, field[0], maxch);
    mkload_date (field[1], start_date, maxch);
    mkload_date (field[2], end_date, maxch);
}
